from pathlib import Path
import logging
import csv

from pydantic import Field
from gpas.models import UploadBase


class UploadData(UploadBase):
    ont_read_suffix: str = Field(
        default=".fastq.gz", description="Suffix for ONT reads"
    )
    illumina_read1_suffix: str = Field(
        default="_1.fastq.gz", description="Suffix for Illumina reads (first of pair)"
    )
    illumina_read2_suffix: str = Field(
        default="_2.fastq.gz", description="Suffix for Illumina reads (second of pair)"
    )
    max_batch_size: int = Field(
        default=50, description="Maximum number of samples per batch"
    )


def build_upload_csv(
    samples_folder: Path | str,
    output_csv: Path | str,
    upload_data: UploadData,
):
    """Create upload csv based on folder of fastq files."""
    samples_folder = Path(samples_folder)
    output_csv = Path(output_csv)
    assert samples_folder.is_dir()  # This should be dealth with by Click

    if upload_data.instrument_platform == "illumina":
        if upload_data.illumina_read1_suffix == upload_data.illumina_read2_suffix:
            raise ValueError("Must have different reads suffixes")

        fastqs1 = list(samples_folder.glob(f"*{upload_data.illumina_read1_suffix}"))
        fastqs2 = list(samples_folder.glob(f"*{upload_data.illumina_read2_suffix}"))

        # sort the lists alphabetically to ensure the files are paired correctly
        fastqs1.sort()
        fastqs2.sort()
        guids1 = [
            f.name.replace(upload_data.illumina_read1_suffix, "") for f in fastqs1
        ]
        guids2 = {
            f.name.replace(upload_data.illumina_read2_suffix, "") for f in fastqs2
        }
        unmatched = guids2.symmetric_difference(guids1)

        if unmatched:
            raise ValueError(
                f"Each sample must have two paired files.\nSome lack pairs:{unmatched}"
            )

        files = [(g, str(f1), str(f2)) for g, f1, f2 in zip(guids1, fastqs1, fastqs2)]
    elif upload_data.instrument_platform == "ont":
        fastqs = list(samples_folder.glob(f"*{upload_data.ont_read_suffix}"))
        fastqs.sort()
        guids = [f.name.replace(upload_data.ont_read_suffix, "") for f in fastqs]
        files = [(g, str(f), str("")) for g, f in zip(guids, fastqs)]
    else:
        raise ValueError("Invalid instrument platform")

    if upload_data.max_batch_size >= len(files):
        _write_csv(
            output_csv,
            files,
            upload_data,
        )
        output_csvs = [output_csv]
    else:
        output_csvs = []
        for i, chunk in enumerate(chunks(files, upload_data.max_batch_size), start=1):
            output_csvs.append(
                output_csv.with_name(f"{output_csv.stem}_{i}{output_csv.suffix}")
            )
            _write_csv(
                output_csv.with_name(f"{output_csv.stem}_{i}{output_csv.suffix}"),
                chunk,
                upload_data,
            )
    logging.info(
        f"Created {len(output_csvs)} CSV files: {', '.join([csv.name for csv in output_csvs])}"
    )
    logging.info("You can use `gpas validate` to check the CSV files before uploading.")


def chunks(lst: list, n: int) -> list[list]:
    """
    Yield successive n-sized chunks from provided list.
    """
    return [lst[i : i + n] for i in range(0, len(lst), n)]


def _write_csv(
    filename: Path,
    read_files: list[tuple[str, str, str]],
    upload_data: UploadData,
):
    """
    Build a CSV file for upload to the Genomic Pathogen Analysis System (GPAS).
    """
    # Note that csv module uses CRLF line endings
    with open(filename, "w", newline="", encoding="utf-8") as outfile:
        csv_writer = csv.writer(outfile)
        csv_writer.writerow(
            [
                "batch_name",
                "sample_name",
                "reads_1",
                "reads_2",
                "control",
                "collection_date",
                "country",
                "subdivision",
                "district",
                "specimen_organism",
                "host_organism",
                "instrument_platform",
            ]
        )
        for sample, f1, f2 in read_files:
            csv_writer.writerow(
                [
                    upload_data.batch_name,
                    sample,
                    f1,
                    f2,
                    "",
                    upload_data.collection_date,
                    upload_data.country,
                    upload_data.subdivision,
                    upload_data.district,
                    upload_data.specimen_organism,
                    upload_data.host_organism,
                    upload_data.instrument_platform,
                ]
            )



import pytest

from gpas.create_upload_csv import UploadData
from datetime import datetime


@pytest.fixture
def upload_data():
    return UploadData(
        batch_name="batch_name",
        instrument_platform="illumina",
        collection_date=datetime.strptime("2024-01-01", "%Y-%m-%d"),
        country="GBR",
        host_organism="homo sapiens",
    )



import csv
import hashlib
import json
import logging
import os
import subprocess
import uuid

from tenacity import retry, wait_random_exponential, stop_after_attempt

from pathlib import Path
from typing import Literal

import httpx


PLATFORMS = Literal["illumina", "ont"]


class UnsupportedClientException(Exception):
    """Exception raised for unsupported client versions"""

    def __init__(self, this_version: str, current_version: str):
        """Raise this exception with a sensible message
        Args:
            this_version (str): The version of installed version
            current_version (str): The version returned by the API
        """
        self.message = (
            f"\n\nThe installed client version ({this_version}) is no longer supported."
            " To update the client, please run:\n\n"
            "conda create -y -n gpas -c conda-forge -c bioconda hostile && conda activate gpas && pip install --upgrade gpas"
        )
        super().__init__(self.message)


# Python errors for neater client errors
class AuthorizationError(Exception):
    """Custom exception for authorization issues. 401"""

    def __init__(self):
        self.message = "Authorization failed! Please re-authenticate with `gpas auth` and try again.\n"
        "If the problem persists please contact support (support@gpas.global)."
        super().__init__(self.message)


class PermissionError(Exception):
    """Custom exception for permission issues. 403"""

    def __init__(self):
        self.message = (
            "You don't have access to this resource! Check logs for more details.\n"
            "Please contact support if you think you should be able to access this resource (support@gpas.global)."
        )
        super().__init__(self.message)


class MissingError(Exception):
    """Custom exception for missing issues. 404"""

    def __init__(self):
        self.message = (
            "Resource not found! It's possible you asked for something which doesn't exist. "
            "Please double check that the resource exists.\n"
            "Note that not all samples have all available output files!"
        )
        super().__init__(self.message)


class ServerSideError(Exception):
    """Custom exception for all other server side errors. 5xx"""

    def __init__(self):
        self.message = (
            "We had some trouble with the server, please double check your command and try again in a moment.\n"
            "If the problem persists, please contact support (support@gpas.global)."
        )
        super().__init__(self.message)


def configure_debug_logging(debug: bool):
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)


def log_request(request):
    logging.debug(f"Request: {request.method} {request.url}")


def log_response(response):
    if response.is_error:
        request = response.request
        response.read()
        logging.error(
            f"{request.method} {request.url} ({response.status_code}) response:\n{json.dumps(response.json(), indent=4)}"
        )


def raise_for_status(response: httpx.Response):
    if response.is_error:
        response.read()
        if response.status_code == 401:
            raise AuthorizationError()
        elif response.status_code == 403:
            raise PermissionError()
        elif response.status_code == 404:
            raise MissingError()
        elif response.status_code // 100 == 5:
            raise ServerSideError()

    # Default to httpx errors in other cases
    response.raise_for_status()


httpx_hooks = {"request": [log_request], "response": [log_response, raise_for_status]}


def run(cmd: str, cwd: Path = Path()):
    return subprocess.run(
        cmd, cwd=cwd, shell=True, check=True, text=True, capture_output=True
    )


def get_access_token(host: str) -> str:
    """Reads token from ~/.config/gpas/tokens/<host>"""
    token_path = Path.home() / ".config" / "gpas" / "tokens" / f"{host}.json"
    logging.debug(f"{token_path=}")
    try:
        data = json.loads((token_path).read_text())
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Token not found at {token_path}, have you authenticated?"
        )
    return data["access_token"].strip()


def parse_csv(csv_path: Path) -> list[dict]:
    """Parse CSV returning a list of dictionaries"""
    with open(csv_path, "r") as fh:
        reader = csv.DictReader(fh)
        return [row for row in reader]


def write_csv(records: list[dict], file_name: Path | str) -> None:
    """Write a list of dictionaries to a CSV file"""
    with open(file_name, "w", newline="") as fh:
        fieldnames = records[0].keys()
        writer = csv.DictWriter(fh, fieldnames=fieldnames)
        writer.writeheader()
        for r in records:
            writer.writerow(r)


def hash_file(file_path: Path) -> str:
    hasher = hashlib.sha256()
    CHUNK_SIZE = 1_048_576  # 2**20, 1MiB
    with open(Path(file_path), "rb") as fh:
        while chunk := fh.read(CHUNK_SIZE):
            hasher.update(chunk)
    return hasher.hexdigest()


@retry(wait=wait_random_exponential(multiplier=2, max=60), stop=stop_after_attempt(10))
def upload_file(
    sample_id: int,
    file_path: Path,
    host: str,
    protocol: str,
    checksum: str,
    dirty_checksum: str,
) -> None:
    with httpx.Client(
        event_hooks=httpx_hooks,
        transport=httpx.HTTPTransport(retries=5),
        timeout=7200,  # 2 hours
    ) as client:
        with open(file_path, "rb") as fh:
            client.post(
                f"{protocol}://{host}/api/v1/samples/{sample_id}/files",
                headers={"Authorization": f"Bearer {get_access_token(host)}"},
                files={"file": fh},
                data={"checksum": checksum, "dirty_checksum": dirty_checksum},
            )


def upload_fastq(
    sample_id: int,
    sample_name: str,
    reads: Path,
    host: str,
    protocol: str,
    dirty_checksum: str,
) -> None:
    """Upload FASTQ file to server"""
    reads = Path(reads)
    logging.debug(f"upload_fastq(): {sample_id=}, {sample_name=}, {reads=}")
    logging.info(f"Uploading {sample_name}")
    checksum = hash_file(reads)
    upload_file(
        sample_id,
        reads,
        host=host,
        protocol=protocol,
        checksum=checksum,
        dirty_checksum=dirty_checksum,
    )
    logging.info(f"  Uploaded {reads.name}")


def upload_paired_fastqs(
    sample_id: int,
    sample_name: str,
    reads_1: Path,
    reads_2: Path,
    host: str,
    protocol: str,
    dirty_checksum_1: str,
    dirty_checksum_2: str,
) -> None:
    """Upload paired FASTQ files to server"""
    reads_1, reads_2 = Path(reads_1), Path(reads_2)
    logging.debug(
        f"upload_paired_fastqs(): {sample_id=}, {sample_name=}, {reads_1=}, {reads_2=}"
    )
    logging.info(f"Uploading {sample_name}")
    checksum1 = hash_file(reads_1)
    checksum2 = hash_file(reads_2)
    upload_file(
        sample_id,
        reads_1,
        host=host,
        protocol=protocol,
        checksum=checksum1,
        dirty_checksum=dirty_checksum_1,
    )
    logging.info(f"  Uploaded {reads_1.name}")
    upload_file(
        sample_id,
        reads_2,
        host=host,
        protocol=protocol,
        checksum=checksum2,
        dirty_checksum=dirty_checksum_2,
    )
    logging.info(f"  Uploaded {reads_2.name}")

    # with concurrent.futures.ThreadPoolExecutor(max_workers=2) as x:
    #     futures = [
    #         x.submit(upload_file, sample_id, reads_1),
    #         x.submit(upload_file, sample_id, reads_2),
    #     ]
    #     for future in concurrent.futures.as_completed(futures):
    #         future.result()


def parse_comma_separated_string(string) -> set[str]:
    return set(string.strip(",").split(","))


def validate_guids(guids: list[str]) -> bool:
    for guid in guids:
        try:
            uuid.UUID(str(guid))
            return True
        except ValueError:
            return False


def map_control_value(v: str) -> bool | None:
    return {"positive": True, "negative": False, "": None}.get(v)


def is_dev_mode() -> bool:
    return True if "GPAS_DEV_MODE" in os.environ else False



import os

import filecmp
import pytest
import logging

from pydantic import ValidationError
from datetime import datetime

from gpas import lib, models
from gpas.util import run
from gpas.create_upload_csv import build_upload_csv, UploadData


def test_cli_version():
    run("gpas --version")


def test_illumina_2():
    lib.upload("tests/data/illumina-2.csv", dry_run=True)
    [os.remove(f) for f in os.listdir(".") if f.endswith("fastq.gz")]
    [os.remove(f) for f in os.listdir(".") if f.endswith(".mapping.csv")]


# # Slow
# def test_ont_2():
#     lib.upload("tests/data/ont-2.csv", dry_run=True)
#     [os.remove(f) for f in os.listdir(".") if f.endswith("fastq.gz")]
#     [os.remove(f) for f in os.listdir(".") if f.endswith(".mapping.csv")]


def test_fail_invalid_fastq_path():
    with pytest.raises(ValidationError):
        lib.upload("tests/data/invalid/invalid-fastq-path.csv", dry_run=True)


def test_fail_empty_sample_name():
    with pytest.raises(ValidationError):
        lib.upload("tests/data/invalid/empty-sample-name.csv", dry_run=True)


def test_fail_invalid_control():
    with pytest.raises(ValidationError):
        lib.upload("tests/data/invalid/invalid-control.csv", dry_run=True)


def test_fail_invalid_specimen_organism():
    with pytest.raises(ValidationError):
        lib.upload("tests/data/invalid/invalid-specimen-organism.csv", dry_run=True)


def test_fail_mixed_instrument_platform():
    with pytest.raises(ValidationError):
        lib.upload("tests/data/invalid/mixed-instrument-platform.csv", dry_run=True)


def test_fail_invalid_instrument_platform():
    with pytest.raises(ValidationError):
        lib.upload("tests/data/invalid/invalid-instrument-platform.csv", dry_run=True)


def test_validate_illumina_model():
    models.parse_upload_csv("tests/data/illumina.csv")
    models.parse_upload_csv("tests/data/illumina-2.csv")


def test_validate_ont_model():
    models.parse_upload_csv("tests/data/ont.csv")


def test_validate_fail_invalid_control():
    with pytest.raises(ValidationError):
        lib.validate("tests/data/invalid/invalid-control.csv")


def test_validate_fail_invalid_specimen_organism():
    with pytest.raises(ValidationError):
        lib.validate("tests/data/invalid/invalid-specimen-organism.csv")


def test_validate_fail_mixed_instrument_platform():
    with pytest.raises(ValidationError):
        lib.validate("tests/data/invalid/mixed-instrument-platform.csv")


def test_validate_fail_invalid_instrument_platform():
    with pytest.raises(ValidationError):
        lib.validate("tests/data/invalid/invalid-instrument-platform.csv")


def test_build_csv_illumina(tmp_path, caplog, upload_data):
    caplog.set_level(logging.INFO)
    build_upload_csv(
        "tests/data/empty_files",
        f"{tmp_path}/output.csv",
        upload_data,
    )

    assert filecmp.cmp(
        "tests/data/auto_upload_csvs/illumina.csv", f"{tmp_path}/output.csv"
    )

    assert "Created 1 CSV files: output.csv" in caplog.text
    assert (
        "You can use `gpas validate` to check the CSV files before uploading."
        in caplog.text
    )


def test_build_csv_ont(tmp_path, caplog, upload_data):
    caplog.set_level(logging.INFO)
    upload_data.instrument_platform = "ont"
    upload_data.district = "dis"
    upload_data.subdivision = "sub"
    upload_data.specimen_organism = "pipe"
    upload_data.host_organism = "unicorn"
    upload_data.ont_read_suffix = "_2.fastq.gz"
    build_upload_csv(
        "tests/data/empty_files",
        f"{tmp_path}/output.csv",
        upload_data,
    )

    assert filecmp.cmp("tests/data/auto_upload_csvs/ont.csv", f"{tmp_path}/output.csv")
    assert "Created 1 CSV files: output.csv" in caplog.text


def test_build_csv_batches(tmp_path, caplog, upload_data):
    caplog.set_level(logging.INFO)
    upload_data.max_batch_size = 3
    build_upload_csv(
        "tests/data/empty_files",
        f"{tmp_path}/output.csv",
        upload_data,
    )

    assert filecmp.cmp(
        "tests/data/auto_upload_csvs/batch1.csv", f"{tmp_path}/output_1.csv"
    )
    assert filecmp.cmp(
        "tests/data/auto_upload_csvs/batch2.csv", f"{tmp_path}/output_2.csv"
    )
    assert "Created 2 CSV files: output_1.csv, output_2.csv" in caplog.text


def test_build_csv_suffix_match(tmp_path, upload_data):
    upload_data.illumina_read2_suffix = "_1.fastq.gz"
    with pytest.raises(ValueError) as e_info:
        build_upload_csv(
            "tests/data/empty_files",
            f"{tmp_path}/output.csv",
            upload_data,
        )
    assert str(e_info.value) == "Must have different reads suffixes"


def test_build_csv_unmatched_files(tmp_path, upload_data):
    with pytest.raises(ValueError) as e_info:
        build_upload_csv(
            "tests/data/unmatched_files",
            f"{tmp_path}/output.csv",
            upload_data,
        )
    assert "Each sample must have two paired files" in str(e_info.value)


def test_build_csv_invalid_tech(tmp_path, upload_data):
    # Note that this should be caught by the model validation
    upload_data.instrument_platform = "invalid"
    with pytest.raises(ValueError) as e_info:
        build_upload_csv(
            "tests/data/unmatched_files",
            f"{tmp_path}/output.csv",
            upload_data,
        )
    assert "Invalid instrument platform" in str(e_info.value)


def test_upload_data_model():
    # Test that making model with invalid country makes error
    with pytest.raises(ValidationError):
        UploadData(
            batch_name="batch_name",
            instrument_platform="invalid",  # type: ignore
            collection_date=datetime.strptime("2024-01-01", "%Y-%m-%d"),
            country="GBR",
        )
    with pytest.raises(ValidationError):
        UploadData(
            batch_name="batch_name",
            instrument_platform="ont",
            collection_date=datetime.strptime("2024-01-01", "%Y-%m-%d"),
            country="G",
        )
    with pytest.raises(ValidationError):
        UploadData(
            batch_name="batch_name",
            instrument_platform="ont",
            collection_date=datetime.strptime("2024-01-01", "%Y-%m-%d"),
            country="GBR",
            specimen_organism="invalid",  # type: ignore
        )



from datetime import date
from pathlib import Path
from typing import Literal

from pydantic import BaseModel, Field, model_validator

from gpas import util


ALLOWED_EXTENSIONS = (".fastq", ".fq", ".fastq.gz", ".fq.gz")


def validate_file_extension(
    filename: str, allowed_extensions: tuple[str] = ALLOWED_EXTENSIONS
):
    return filename.endswith(allowed_extensions)


class UploadBase(BaseModel):
    batch_name: str = Field(
        default=None, description="Batch name (anonymised prior to upload)"
    )
    instrument_platform: util.PLATFORMS = Field(
        description="Sequencing instrument platform"
    )
    collection_date: date = Field(description="Collection date in yyyy-mm-dd format")
    country: str = Field(
        min_length=3, max_length=3, description="ISO 3166-2 alpha-3 country code"
    )
    subdivision: str = Field(
        default=None, description="ISO 3166-2 principal subdivision"
    )
    district: str = Field(default=None, description="Granular location")
    specimen_organism: Literal["mycobacteria", ""] = Field(
        default="mycobacteria", description="Target specimen organism scientific name"
    )
    host_organism: str = Field(
        default=None, description="Host organism scientific name"
    )


class UploadSample(UploadBase):
    sample_name: str = Field(
        min_length=1, description="Sample name (anonymised prior to upload)"
    )
    upload_csv: Path = Field(description="Absolute path of upload CSV file")
    reads_1: Path = Field(description="Relative path of first FASTQ file")
    reads_2: Path = Field(
        description="Relative path of second FASTQ file", default=None
    )
    control: Literal["positive", "negative", ""] = Field(
        description="Control status of sample"
    )

    @model_validator(mode="after")
    def check_fastqs_are_different(self):
        if self.reads_1 == self.reads_2:
            raise ValueError("reads_1 and reads_2 paths must be different")
        return self

    @model_validator(mode="after")
    def validate_fastqs_by_platform(self):
        reads_1 = self.reads_1
        reads_2 = self.reads_2
        reads_1_resolved_path = self.upload_csv.resolve().parent / reads_1
        reads_2_resolved_path = self.upload_csv.resolve().parent / reads_2
        if self.instrument_platform == "ont":
            if not reads_1_resolved_path.is_file():
                raise ValueError("reads_1 must be a valid FASTQ file path")
            if reads_2_resolved_path.is_file():
                raise ValueError(
                    "reads_2 must be empty where instrument_platform is ont"
                )
            if reads_1 and not validate_file_extension(reads_1.name):
                raise ValueError(
                    f"Invalid file extension for file {self.reads_1.name}. Allowed extensions are {ALLOWED_EXTENSIONS}"
                )
        elif self.instrument_platform == "illumina":
            if (
                not reads_1_resolved_path.is_file()
                or not reads_2_resolved_path.is_file()
            ):
                raise ValueError(
                    "reads_1 and reads_2 must be valid FASTQ file paths where instrument_platform is illumina"
                )
            if reads_1 and not validate_file_extension(reads_1.name):
                raise ValueError(
                    f"Invalid file extension for file {self.reads_1.name}. Allowed extensions are {ALLOWED_EXTENSIONS}"
                )
            if reads_2 and not validate_file_extension(reads_2.name):
                raise ValueError(
                    f"Invalid file extension for file {self.reads_2.name}. Allowed extensions are {ALLOWED_EXTENSIONS}"
                )
        return self

    # @model_validator(mode="after")
    # def enforce_dev_mode(self):
    #     dev_mode = util.is_dev_mode()  # Use this bool for client side feature gating
    #     return self

    # @model_validator(pre=True)
    # def lowercase_all_fields(cls, values: dict[str, Any]) -> dict[str, Any]:
    #     return {k: (v.lower() if isinstance(v, str) else v) for k, v in values.items()}


class UploadBatch(BaseModel):
    samples: list[UploadSample]

    @model_validator(mode="after")
    def validate_unique_sample_names(self):
        names = [sample.sample_name for sample in self.samples]
        if len(names) != len(set(names)):
            raise ValueError("Found duplicate sample names")
        return self

    @model_validator(mode="after")
    def validate_unique_file_names(self):
        reads_1_filenames = [str(sample.reads_1.name) for sample in self.samples]
        reads_2_filenames = [str(sample.reads_2.name) for sample in self.samples]
        instrument_platforms = self.samples[0].instrument_platform
        if instrument_platforms == "ont":
            if len(reads_1_filenames) != len(set(reads_1_filenames)):
                raise ValueError("Found duplicate FASTQ filenames")
        elif instrument_platforms == "illumina":
            if len(reads_1_filenames) + len(reads_1_filenames) != len(
                set(reads_1_filenames) | set(reads_2_filenames)
            ):
                raise ValueError("Found duplicate FASTQ filenames")
        return self

    @model_validator(mode="after")
    def validate_single_instrument_platform(self):
        instrument_platforms = [sample.instrument_platform for sample in self.samples]
        if len(set(instrument_platforms)) != 1:
            raise ValueError(
                "Samples within a batch must have the same instrument_platform"
            )
        return self


class RemoteFile(BaseModel):
    filename: str
    run_id: int
    sample_id: str


def parse_upload_csv(upload_csv: Path) -> UploadBatch:
    records = util.parse_csv(upload_csv)
    return UploadBatch(  # Include upload_csv to enable relative fastq path validation
        samples=[UploadSample(**r, **dict(upload_csv=upload_csv)) for r in records]
    )



from datetime import datetime, date
import json as json_
from getpass import getpass
from pathlib import Path

import click

from gpas import lib, util
from gpas.create_upload_csv import build_upload_csv, UploadData


@click.group(name="GPAS")
@click.version_option()
@click.help_option("-h", "--help")
def main():
    """GPAS command line interface."""
    pass


@main.command()
@click.option("--host", type=str, default=None, help="API hostname (for development)")
def auth(
    *,
    host: str | None = None,
) -> None:
    """
    Authenticate with the GPAS platform.
    """
    host = lib.get_host(host)
    username = input("Enter your username: ")
    password = getpass(prompt="Enter your password: ")
    lib.authenticate(username=username, password=password, host=host)


@main.command()
@click.argument(
    "upload_csv", type=click.Path(exists=True, dir_okay=False, readable=True)
)
@click.option(
    "--threads",
    type=int,
    default=None,
    help="Number of alignment threads used during decontamination",
)
@click.option(
    "--save", is_flag=True, help="Retain decontaminated reads after upload completion"
)
@click.option("--dry-run", is_flag=True, help="Exit before uploading reads")
@click.option("--host", type=str, default=None, help="API hostname (for development)")
@click.option("--debug", is_flag=True, help="Enable verbose debug messages")
def upload(
    upload_csv: Path,
    *,
    # out_dir: Path = Path(),
    threads: int | None = None,
    save: bool = False,
    dry_run: bool = False,
    host: str | None = None,
    debug: bool = False,
) -> None:
    """
    Validate, decontaminate and upload reads to the GPAS platform. Creates a mapping CSV
    file which can be used to download output files with original sample names.
    """
    # :arg out_dir: Path of directory in which to save mapping CSV
    util.configure_debug_logging(debug)
    host = lib.get_host(host)
    lib.upload(upload_csv, save=save, dry_run=dry_run, threads=threads, host=host)


@main.command()
@click.argument("samples", type=str)
@click.option(
    "--filenames",
    type=str,
    default="main_report.json",
    help="Comma-separated list of output filenames to download",
)
@click.option(
    "--inputs", is_flag=True, help="Also download decontaminated input FASTQ file(s)"
)
@click.option(
    "--out-dir", type=click.Path(file_okay=False), default=".", help="Output directory"
)
@click.option(
    "--rename/--no-rename",
    default=True,
    help="Rename downloaded files using sample names when given a mapping CSV",
)
@click.option("--host", type=str, default=None, help="API hostname (for development)")
@click.option("--debug", is_flag=True, help="Enable verbose debug messages")
def download(
    samples: str,
    *,
    filenames: str = "main_report.json",
    inputs: bool = False,
    out_dir: Path = Path(),
    rename: bool = True,
    host: str | None = None,
    debug: bool = False,
) -> None:
    """
    Download input and output files associated with sample IDs or a mapping CSV file
    created during upload.
    """
    util.configure_debug_logging(debug)
    host = lib.get_host(host)
    if util.validate_guids(util.parse_comma_separated_string(samples)):
        lib.download(
            samples=samples,
            filenames=filenames,
            inputs=inputs,
            out_dir=out_dir,
            host=host,
        )
    elif Path(samples).is_file():
        lib.download(
            mapping_csv=samples,
            filenames=filenames,
            inputs=inputs,
            out_dir=out_dir,
            rename=rename,
            host=host,
        )
    else:
        raise ValueError(
            f"{samples} is neither a valid mapping CSV path nor a comma-separated list of valid GUIDs"
        )


@main.command()
@click.argument("samples", type=str)
@click.option("--host", type=str, default=None, help="API hostname (for development)")
@click.option("--debug", is_flag=True, help="Enable verbose debug messages")
def query_raw(samples: str, *, host: str | None = None, debug: bool = False) -> None:
    """
    Fetch metadata for one or more SAMPLES in JSON format.
    SAMPLES should be command separated list of GUIDs or path to mapping CSV.
    """
    util.configure_debug_logging(debug)
    host = lib.get_host(host)
    if util.validate_guids(util.parse_comma_separated_string(samples)):
        result = lib.query(samples=samples, host=host)
    elif Path(samples).is_file():
        result = lib.query(mapping_csv=samples, host=host)
    else:
        raise ValueError(
            f"{samples} is neither a valid mapping CSV path nor a comma-separated list of valid GUIDs"
        )
    print(json_.dumps(result, indent=4))


@main.command()
@click.argument("samples", type=str)
@click.option("--json", is_flag=True, help="Output status in JSON format")
@click.option("--host", type=str, default=None, help="API hostname (for development)")
@click.option("--debug", is_flag=True, help="Enable verbose debug messages")
def query_status(
    samples: str, *, json: bool = False, host: str | None = None, debug: bool = False
) -> None:
    """
    Fetch processing status for one or more SAMPLES.
    SAMPLES should be command separated list of GUIDs or path to mapping CSV.
    """
    util.configure_debug_logging(debug)
    host = lib.get_host(host)
    if util.validate_guids(util.parse_comma_separated_string(samples)):
        result = lib.status(samples=samples, host=host)
    elif Path(samples).is_file():
        result = lib.status(mapping_csv=samples, host=host)
    else:
        raise ValueError(
            f"{samples} is neither a valid mapping CSV path nor a comma-separated list of valid GUIDs"
        )
    if json:
        print(json_.dumps(result, indent=4))
    else:
        for name, status in result.items():
            print(f"{name} \t{status}")


@main.command()
def download_index() -> None:
    """
    Download and cache host decontamination index.
    """
    lib.download_index()


@main.command()
@click.argument(
    "upload_csv", type=click.Path(exists=True, dir_okay=False, readable=True)
)
@click.option("--host", type=str, default=None, help="API hostname (for development)")
@click.option("--debug", is_flag=True, help="Enable verbose debug messages")
def validate(upload_csv: Path, *, host: str | None = None, debug: bool = False) -> None:
    """Validate a given upload CSV."""
    util.configure_debug_logging(debug)
    host = lib.get_host(host)
    lib.validate(upload_csv, host=host)


# In future this could be updated based on a file
defaults = {
    "country": None,
    "district": "",
    "subdivision": "",
    "instrument_platform": "illumina",
    "ont_read_suffix": ".fastq.gz",
    "illumina_read1_suffix": "_1.fastq.gz",
    "illumina_read2_suffix": "_2.fastq.gz",
    "max_batch_size": 50,
}


@main.command()
@click.argument(
    "samples-folder", type=click.Path(exists=True, file_okay=False), required=True
)
@click.option(
    "--output-csv",
    type=click.Path(dir_okay=False),
    default="upload.csv",
    help="Path to output CSV file",
    required=True,
)
@click.option("--batch-name", type=str, help="Batch name", required=True)
@click.option(
    "--collection-date",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=str(date.today()),
    show_default=True,
    help="Collection date (YYYY-MM-DD)",
    required=True,
)
@click.option(
    "--country",
    type=str,
    help="3-letter Country Code",
    required=True,
    default=defaults["country"],
    show_default=True,
)
@click.option(
    "--instrument-platform",
    type=click.Choice(["illumina", "ont"]),
    default=defaults["instrument_platform"],
    help="Sequencing technology",
)
@click.option(
    "--subdivision",
    type=str,
    help="Subdivision",
    default=defaults["subdivision"],
    show_default=True,
)
@click.option(
    "--district",
    type=str,
    help="District",
    default=defaults["district"],
    show_default=True,
)
@click.option(
    "--ont_read_suffix",
    type=str,
    default=defaults["ont_read_suffix"],
    help="Read file ending for ONT fastq files",
    show_default=True,
)
@click.option(
    "--illumina_read1_suffix",
    type=str,
    default=defaults["illumina_read1_suffix"],
    help="Read file ending for Illumina read 1 files",
    show_default=True,
)
@click.option(
    "--illumina_read2_suffix",
    type=str,
    default=defaults["illumina_read2_suffix"],
    help="Read file ending for Illumina read 2 files",
    show_default=True,
)
@click.option("--max-batch-size", type=int, default=50, show_default=True)
def build_csv(
    samples_folder: Path,
    output_csv: Path,
    instrument_platform: str,
    batch_name: str,
    collection_date: datetime,
    country: str,
    subdivision: str = "",
    district: str = "",
    pipeline: str = "mycobacteria",
    host_organism: str = "homo sapiens",
    ont_read_suffix: str = ".fastq.gz",
    illumina_read1_suffix: str = "_1.fastq.gz",
    illumina_read2_suffix: str = "_2.fastq.gz",
    max_batch_size: int = 50,
):
    """
    Command to create upload csv from SAMPLES_FOLDER containing sample fastqs.\n
    Use max_batch_size to split into multiple separate upload csvs.\n
    Adjust the read_suffix parameters to match the file endings for your read files.
    """
    if len(country) != 3:
        raise ValueError(f"Country ({country}) should be 3 letter code")
    output_csv = Path(output_csv)
    samples_folder = Path(samples_folder)

    upload_data = UploadData(
        batch_name=batch_name,
        instrument_platform=instrument_platform,  # type: ignore
        collection_date=collection_date,
        country=country,
        subdivision=subdivision,
        district=district,
        specimen_organism=pipeline,  # type: ignore
        host_organism=host_organism,
        ont_read_suffix=ont_read_suffix,
        illumina_read1_suffix=illumina_read1_suffix,
        illumina_read2_suffix=illumina_read2_suffix,
        max_batch_size=max_batch_size,
    )

    build_upload_csv(
        samples_folder,
        output_csv,
        upload_data,
    )
